"""
Actionista Action CLI for Todoist (actionista-todoist).

A `find`-inspired CLI for Todoist.

Add, select, print, reschedule, modify, and complete tasks in a batch-wise fashion
from the command line.

"""

# Do not zero-pad month or day:
__version__ = '2019.9.10'  # remember to also update setup.py
